/*

  route-fix.c: a program to fix bogus classful routes on android phones
  (c) 2010 Patrick Cernko <pcernko@mpi-sws.org>

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/route.h>
#include <sys/ioctl.h>

#ifndef RTF_UP
/* Keep this in sync with /usr/src/linux/include/linux/route.h */
#define RTF_UP          0x0001	/* route usable                 */
#define RTF_GATEWAY     0x0002	/* destination is a gateway     */
#define RTF_HOST        0x0004	/* host entry (net otherwise)   */
#define RTF_REINSTATE   0x0008	/* reinstate route after tmout  */
#define RTF_DYNAMIC     0x0010	/* created dyn. (by redirect)   */
#define RTF_MODIFIED    0x0020	/* modified dyn. (by redirect)  */
#define RTF_MTU         0x0040	/* specific MTU for this route  */
#ifndef RTF_MSS
#define RTF_MSS         RTF_MTU	/* Compatibility :-(            */
#endif
#define RTF_WINDOW      0x0080	/* per route window clamping    */
#define RTF_IRTT        0x0100	/* Initial round trip time      */
#define RTF_REJECT      0x0200	/* Reject route                 */
#endif

static const unsigned flagvals[] = { /* Must agree with flagchars[]. */
	RTF_GATEWAY,
	RTF_HOST,
	RTF_REINSTATE,
	RTF_DYNAMIC,
	RTF_MODIFIED,
#if ENABLE_FEATURE_IPV6
	RTF_DEFAULT,
	RTF_ADDRCONF,
	RTF_CACHE
#endif
};

#define IPV4_MASK (RTF_GATEWAY|RTF_HOST|RTF_REINSTATE|RTF_DYNAMIC|RTF_MODIFIED)
#define IPV6_MASK (RTF_GATEWAY|RTF_HOST|RTF_DEFAULT|RTF_ADDRCONF|RTF_CACHE)

/* Must agree with flagvals[]. */
static const char flagchars[] =
	"GHRDM"
#if ENABLE_FEATURE_IPV6
	"DAC"
#endif
;

struct routelist_t;
struct routelist_t {
  struct routelist_t *next;
  struct routelist_t *prev;
  unsigned long dest;
  unsigned long gw;
  unsigned long mask;
  char dev[64];
};

static void set_flags(char *flagstr, int flags)
{
  int i;

  *flagstr++ = 'U';

  for (i = 0; (*flagstr = flagchars[i]) != 0; i++) {
    if (flags & flagvals[i]) {
      ++flagstr;
    }
  }
}

struct routelist_t* get_routes() {
  char devname[64], flags[16];
  unsigned long d, g, m;
  int flgs, ref, use, metric, mtu, win, ir;
  struct routelist_t* routes = NULL;
  struct routelist_t* curroute = NULL;
  struct routelist_t* newroute = NULL;

  FILE *fp = fopen("/proc/net/route", "r");

  if (fscanf(fp, "%*[^\n]\n") < 0) { /* Skip the first line. */
    exit(1);		   /* Empty or missing line, or read error. */
  }
  while (1) {
    int r;
    r = fscanf(fp, "%63s%lx%lx%X%d%d%d%lx%d%d%d\n",
	       devname, &d, &g, &flgs, &ref, &use, &metric, &m,
	       &mtu, &win, &ir);
    if (r != 11) {
      if ((r < 0) && feof(fp)) { /* EOF with no (nonspace) chars read. */
	break;
      }
      exit(2);
    }

    if (!(flgs & RTF_UP)) { /* Skip interfaces that are down. */
      continue;
    }

    set_flags(flags, (flgs & IPV4_MASK));
#ifdef RTF_REJECT
    if (flgs & RTF_REJECT) {
      flags[0] = '!';
    }
#endif

    newroute = malloc(sizeof(struct routelist_t));
    if (routes) {
      curroute->next = newroute;
      newroute->prev = curroute;
    } else {
      newroute->prev = NULL;
      routes = newroute;
    }
    newroute->next = NULL;
    newroute->dest = d;
    newroute->gw = g;
    newroute->mask = m;
    strncpy(newroute->dev, devname, 64);
    curroute = newroute;
  }

  return routes;
}

int main(int argc, char* argv[]) {
  struct routelist_t* routes = get_routes();
  struct routelist_t* r = routes;
  unsigned int verbose = argc - 1;

  while (r) {
    struct in_addr m;
    m.s_addr = r->mask;
    if (verbose > 1) printf("0x%08lx 0x%08lx 0x%08lx %s\n", r->dest, r->gw, r->mask, r->dev);
    if (r->gw == 0) {
      /* direct route found, lets see if it is a default class A/B route */
      if ( ((r->mask == 0x000000ff) && (r->dest < 0x08000000)) ||  /* CLASS B network and corresponding mask */
	   ((r->mask == 0x0000ffff) && (r->dest < 0xC0000000)) ) { /* CLASS A network and corresponding mask */
	int sock = socket(AF_INET, SOCK_DGRAM, 0);
	struct rtentry rt;
	struct sockaddr_in* work;
	if (verbose) printf("Found bogus CLASS %c default route: 0x%08lx 0x%08lx 0x%08lx %s\n",
			    ((r->mask == 0x0000ffff) ? 'B' : 'A'),
			    r->dest, r->gw, r->mask, r->dev);	
	memset(&rt, 0, sizeof(rt));
	work = (struct sockaddr_in*) &rt.rt_dst;
	work->sin_family = AF_INET;
	work->sin_port = 0;
	work->sin_addr.s_addr = r->dest;
	work = (struct sockaddr_in*) &rt.rt_gateway;
	work->sin_family = AF_INET;
	work->sin_port = 0;
	work->sin_addr.s_addr = r->gw;
	work = (struct sockaddr_in*) &rt.rt_genmask;
	work->sin_family = AF_INET;
	work->sin_port = 0;
	work->sin_addr.s_addr = r->mask;
	if (ioctl(sock, SIOCDELRT, &rt)) {
	  perror("Failed to remove bogus CLASS A/B default route: ");
	}
      }
    }	 
    r = r->next;
  }
  return 0;
}
